# Python data structures for the ast nodes
# The traverse method is just a method to
# print out the AST and it shows some ideas
# in how the traverse of the AST can be easily
# done. But for type checks and code generation,
# we should apply the visitor pattern.

class Py_OclExpression:
    '''
    An abstract class for OclExpression
    '''
    def __init__(self):
        raise RuntimeError, "OclExpression can't be instantiated"

class Py_VariableExp(Py_OclExpression):
    '''
    VariableExp inherite from OclExpression
    '''
    def __init__(self, name):
        '''
        name: Py_Name
        '''
        self.type = 'VARIABLEEXP'
        self.name = name

    def traverse(self):
        return self.name.traverse()

class Py_Name:
    '''
    A linked list containing the whole path
    '''
    def __init__(self, value, next):
        '''
        value: String
        next: Py_Name
        '''
        self.type = 'NAME'
        self.value = value  # string value of the name
        self.next = next

    def traverse(self):
        str = self.value
        if (self.next == None):
            return str
        str += '::' + self.next.traverse()
        return str
            

class Py_LiteralExp(Py_OclExpression):
    '''
    An abstract class for LiteralExp, inherites from OclExpression
    '''
    def __init__(self):
      raise RuntimeError, "LiteralExp can't be instantiated"

class Py_EnumLiteralExp(Py_LiteralExp):
    '''
    EnumLiteralExp inherites from LiteralExp
    '''
    def __init__(self, name, value):
        '''
        enum: Py_Name
        value: String
        '''
        self.type = 'ENUMLITERALEXP'
        self.enum = name
        self.value = value

    def traverse(self):
        str = self.enum.traverse() + ':::' + self.value
        return str

class Py_CollectionLiteralExp(Py_LiteralExp):
    '''
    CollectionLiteralExp inherites from LiteralExp
    Three types of CollectionLiteral:
    Sequence{1,2}, Bag{1,2}, Set{1,2}
    '''
    def __init__(self, type, items):
        '''
        type: String
        items: Py_CollectionItem
        '''
        self.type = 'COLLECTIONLITERALEXP'
        self.collectionType = type  # can only be 'Set', 'Bag', 'Sequence'
        self.collectionItem = items

    def traverse(self):
        str = self.collectionType
        if (self.collectionItem == None):
            str += '{}'
            return str
        str += '{' + self.collectionItem.traverse() + '}'
        return str

class Py_CollectionItem:
    '''
    A linked list of collection items
    '''
    def __init__(self, currentItem, next):
        '''
        individualItem: can be either a Py_CollectionRange
                        or a Py_OclExpression
        next: Py_CollectionItem
        '''
        self.type = 'COLLECTIONITEM'
        self.individualItem = currentItem 
        self.next = next

    def traverse(self):
        str = self.individualItem.traverse()
        if (self.next != None):
            str += ', ' + self.next.traverse()
        return str

class Py_CollectionRange:
    '''
    CollectionRange,
    shorthand for defining a sequence of intergers
    '''
    def __init__(self, first, last):
        '''
        first: Py_OclExpression
        last: Py_OclExpression
        '''
        self.type = 'COLLECTIONRANGE'
        self.first = first  # first is an OclExpression
        self.last = last    # last is an OclExpression

    def traverse(self):
        str = self.first.traverse()
        str += '..' + self.last.traverse()
        return str

class Py_NumericLiteralExp(Py_LiteralExp):
    '''
    An abstract class for NumericLiteralExp
    inherites from LiteralExp
    can be either a real or an integer
    '''
    def __init__(self):
        raise RuntimeError, "NumericLiteralExp can't be instantiated"
        
class Py_IntegerLiteralExp(Py_NumericLiteralExp):
    '''
    IntegerLiteralExp inherites from NumericLiteralExp
    '''
    def __init__(self, value):
        self.type = 'INTEGERLITERALEXP'
        self.value = value  # value should be an integer

    def traverse(self):
        return str(self.value)

class Py_RealLiteralExp(Py_NumericLiteralExp):
    '''
    RealLiteralExp inherites from NumericLiteralExp
    '''
    def __init__(self, value):
        self.type = 'REALLITERALEXP'
        self.value = value  # value should be an real

    def traverse(self):
        return str(self.value)

class Py_StringLiteralExp(Py_LiteralExp):
    '''
    StringLiteralExp inherites from LiteralExp
    '''
    def __init__(self, value):
        '''
        value should be either 1(true) or 0(false)
        '''
        self.type = 'STRINGLITERALEXP'
        self.value = value  # value should be a string

    def traverse(self):
        return self.value
        
class Py_BooleanLiteralExp(Py_LiteralExp):
    '''
    BooleanLiteralExp inherites from LiteralExp
    '''
    def __init__(self, value):
        self.type = 'BOOLEANLITERALEXP'
        self.value = value  # value should be a boolean

    def traverse(self):
        if (self.value == 1):
            return 'true'
        return 'false'

class Py_TupleLiteralExp(Py_LiteralExp):
    '''
    TupleLiteralExp inherites from LiteralExp
    '''
    def __init__(self, parts):
        self.type = 'TUPLELITERALEXP'
        self.tupleLiteralExpParts = parts

    def traverse(self):
        str = 'Tuple'
        if (self.tupleLiteralExpParts == None):
            str += '{}'
        else:
            str += '{' + self.tupleLiteralExpParts.traverse() + '}'
        return str

class Py_TupleLiteralExpParts:
    '''
    TupleLiteralExpParts
    '''
    def __init__(self, head, tail):
        self.head = head  # head should be a VariableDeclaration
        self.tail = tail  # tail should be a TupleLiteralExpParts

    def traverse(self):
        str = self.head.traverse()
        if (self.tail != None):
            str += ', ' + self.tail.traverse()
        return str

class Py_PropertyCallExp(Py_OclExpression):
    '''
    An abstract class for PropertyCallExp
    inherites from OclExpression
    '''
    def __init__(self):
        raise RuntimeError, "PropertyCallExp can't be instantiated"

class Py_LoopExp(Py_PropertyCallExp):
    '''
    An abstract class for LoopExp inherites from PropertyCallExp
    '''
    def __init__(self):
        raise RuntimeError, "LoopExp can't be instantiated"

class Py_IteratorExp(Py_LoopExp):
    '''
    IteratorExp inherites from LoopExp
    '''
    def __init__(self, src, body, iterator, op):
        '''
        src: Py_OclExpression
        body: Py_OclExpression
        iterator: Py_VariableDeclaration
        op: String
        '''
        self.type = 'ITERATOREXP'
        self.source = src  
        self.body = body   
        self.iterator = iterator  
        # op should be one of the following:
        # 'SELECT', 'COLLECT', 'REJECT', 'FORALL', 'EXISTS',
        # 'ISUNIQUE', 'SIZE', ISEMPTY', 'NOTEMPTY', 'COUNT',
        # 'UNION', 'DIFFERENCE'
        self.op = op

    def traverse(self):
        str = self.source.traverse()
        str += '->' + self.op + '('
        if (self.iterator != None):
            str += self.iterator.traverse()
        str += ' | ' + self.body.traverse() + ')'
        return str

class Py_IterateExp(Py_LoopExp):
    '''
    IterateExp inherites from LoopExp
    '''
    def __init__(self, src, body, iterator, result):
        self.type = 'ITERATEEXP'
        self.source = src  # sourc is an OclExpression
        self.body = body   # body is an OclExpression
        self.iterator = iterator  # iterator is a VariableDeclaration
        self.result = result      # result is a VariableDeclaration

    def traverse(self):
        str = self.source.traverse()
        str += '->' + 'iterate' + '('
        if (self.iterator != None):
            str += self.iterator.traverse()
        str += '; ' + self.result.traverse() + ' | '
        str += self.body.traverse() + ')'
        return str


class Py_VariableDeclaration:
    '''
    VariableDeclaration
    '''
    def __init__(self, name, type, init):
        '''
        name: String
        type: Py_Name
        init: Py_OclExpression
        '''
        self.type = 'VARIABLEDECLARATION'
        self.varName = name  # name is a string
        self.varType = type  # type is a Py_Name
        self.initExpression = init

    def traverse(self):
        str = self.varName
        if (self.varType != None):
            str += ':' + self.varType.traverse()
        if (self.initExpression != None):
            str += '=' + self.initExpression.traverse()
        return str

class Py_ModelPropertyCallExp(Py_PropertyCallExp):
    '''
    An abstract class for ModelPropertyCallExp
    '''
    def __init__(self):
        raise RuntimeError, "ModelPropertyCallExp can't be instantiated"

class Py_OperationCallExp(Py_ModelPropertyCallExp):
    '''
    OperationCallExp
    '''
    def __init__(self, src, args, op, timeExpr):
        '''
        src: Py_OclExpression
        args: Py_Arguments
        op: String
        timeExpr: Boolean
        '''
        self.type = 'OPERATIONCALLEXP'
        self.source = src
        self.arguments = args
        self.op = op
        self.timeExpr = timeExpr

    def traverse(self):
        '''
        str: String, store Ocl expression
        '''
        # print out binary operation
        L = ["+","-","*","/","and","or","xor",\
             "=","<>","<=",">=","<",">",\
             "implies"]
        for op in L:
            if (self.op == op):
                str = '(' + self.source.traverse() + ' ' + op
                str += ' ' + self.arguments.traverse() + ')'
                return str
                
        ## if (self.op == 'PLUS'):
##             str = '(' + self.source.traverse()
##             str += ' + ' + self.arguments.traverse() + ')'
##         elif (self.op == 'MINUS'):
##             str = '(' + self.source.traverse()
##             str += ' - ' + self.arguments.traverse() + ')'
##         elif (self.op == 'TIMES'):
##             str = '(' + self.source.traverse()
##             str += ' * ' + self.arguments.traverse() + ')'
##         elif (self.op == 'DIVIDES'):
##             str = '(' + self.source.traverse()
##             str += ' / ' + self.arguments.traverse() + ')'
##         elif (self.op == 'AND'):
##             str = '(' + self.source.traverse()
##             str += ' AND ' + self.arguments.traverse() + ')'
##         elif (self.op == 'OR'):
##             str = '(' + self.source.traverse()
##             str += ' OR ' + self.arguments.traverse() + ')'
##         elif (self.op == 'XOR'):
##             str = '(' + self.source.traverse()
##             str += ' XOR ' + self.arguments.traverse() + ')'
##         elif (self.op == 'EQ'):
##             str = '(' + self.source.traverse()
##             str += ' = ' + self.arguments.traverse() + ')'
##         elif (self.op == 'NOTEQ'):
##             str = '(' + self.source.traverse()
##             str += ' <> ' + self.arguments.traverse() + ')'
##         elif (self.op == 'LTEQ'):
##             str = '(' + self.source.traverse()
##             str += ' <= ' + self.arguments.traverse() + ')'
##         elif (self.op == 'GTEQ'):
##             str = '(' + self.source.traverse()
##             str += ' >= ' + self.arguments.traverse() + ')'
##         elif (self.op == 'LT'):
##             str = '(' + self.source.traverse()
##             str += ' < ' + self.arguments.traverse() + ')'
##         elif (self.op == 'GT'):
##             str = '(' + self.source.traverse()
##             str += ' > ' + self.arguments.traverse() + ')'
##         elif (self.op == 'IMPLIES'):
##             str = '(' + self.source.traverse()
##             str += ' IMPLIES ' + self.arguments.traverse() + ')'

        # print out ocl standard library collection related operation
        L2 = ['size','includes','excludes','count','includesAll',\
              'excludesAll','isEmpty','notEmpty','sum','union',\
              'intersection','including','excluding',\
              'symmetricDifference','flatten','asSet','asBag',\
              'asSequence','append','prepend','subSequence',\
              'first','at','last']
        for op in L2:
            if (self.op == op):                   
                str = self.source.traverse() + '->' + self.op
                if (self.timeExpr == 1):
                    str += '@pre'
                if (self.arguments == None):
                    str += '()'
                else:
                    str += '(' + self.arguments.traverse() + ')'
                return str

        # print out "NOT" operation
        if (self.op == 'NOT'):
            return '(not ' + self.source.traverse() + ')'
        else: # print out UML model operation                  
            str = self.source.traverse() + '.' + self.op
            if (self.timeExpr == 1):
                str += '@pre'
            if (self.arguments == None):
                str += '()'
            else:
                str += '(' + self.arguments.traverse() + ')'
            return str

class Py_AttrOrNavCallExp(Py_ModelPropertyCallExp):
    '''
    An AttrOrNavCallExp could be either an AttributeCallExp
    or a NavigationCallExp
    But this would not be know without accessing the UML model
    '''
    def __init__(self, src, attrOrAss, timeExpr):
        '''
        src: Py_OclExpression
        attrOrAss: String, attribute or association
        timeExpr: Boolean
        '''
        self.type = 'ATTRORNAVCALLEXP'
        self.source = src
        self.attrOrAss = attrOrAss
        self.timeExpr = timeExpr

    def traverse(self):
        '''
        str: String, store Ocl expression
        '''
        str = self.source.traverse() + '.' + self.attrOrAss
        if (self.timeExpr == 1):
            str += '@pre'
        return str
    

class Py_Arguments:
    '''
    A linked list of arguments
    '''
    def __init__(self, head, tail):
        '''
        head: Py_OclExpression
        tail: Py_Arguments
        '''
        self.type = 'ARGUMENTS'
        self.head = head
        self.tail = tail

    def traverse(self):
        if (self.head == None):
            return ""
        str = self.head.traverse()
        if (self.tail != None):
            str += ', ' + self.tail.traverse()
        return str

class Py_LetExp(Py_OclExpression):
    '''
    LetExp
    '''
    def __init__(self, vars, letExpSub):
        '''
        vars: Py_Variables
        letExpSub: Py_OclExpression
        '''
        self.type = 'LETEXP'
        self.vars = vars
        self.letExpSub = letExpSub

    def traverse(self):
        str = 'Let ' + self.vars.traverse()
        str += ' in ( \n' + self.letExpSub.traverse() + ')'
        return str

class Py_Variables:
    '''
    Variables
    '''
    def __init__(self, head, tail):
        '''
        head: Py_VariableDeclaration
        tail: Py_Variables
        '''
        self.type = 'VARIABLES'
        self.head = head
        self.tail = tail

    def traverse(self):
        str = self.head.traverse()
        if (self.tail != None):
            str += ';\n    ' + self.tail.traverse()
        return str

class Py_OclMessageExp(Py_OclExpression):
    '''
    OclMessageExp
    '''
    def __init__(self, target, actionOrSignal, arg):
        '''
        target: Py_OclExpression
        actionOrSignal: String
        arg: Py_OclMessageArg
        '''
        self.type = 'OCLMESSAGEEXP'
        self.target = target
        self.actionOrSignal = actionOrSignal
        self.oclMessageArg = arg

    def traverse(self):
        str = self.target.traverse() + '^' + self.actionOrSignal
        if (self.oclMessageArg == None):
            str += '()'
        else:
            str += '(' + self.oclMessageArg.traverse() + ')'
        return str

class Py_OclMessageArg:
    '''
    OclMessageArg
    '''
    def __init__(self, expr, next):
        '''
        expr could be an OclExpression or an UnspecifiedValueExp
        next: Py_OclMessageArg
        '''
        self.type = 'OCLMESSAGEARG'
        self.expr = expr
        self.next = next

    def traverse(self):
        str = self.expr.traverse()
        if (self.next != None):
            str += ', ' + self.next.traverse()
        return str
        

class Py_UnspecifiedValueExp:
    '''
    UnspecifiedValueExp
    '''
    def __init__(self, varType):
        '''
        varType: Py_Name
        '''
        self.type = 'UNSPECIFIEDVALUEEXP'
        self.varType = varType

    def traverse(self):
        if (self.varType == None):
            return '?'
        return '?:' + self.varType.traverse()

class Py_IfExp(Py_OclExpression):
    '''
    IfExp
    '''
    def __init__(self, cond, thenExp, elseExp):
        '''
        cond: Py_OclExpression
        thenExp: Py_OclExpression
        elseExp: Py_OclExpression
        '''
        self.type = 'IFEXP'
        self.condition = cond
        self.thenExpression = thenExp
        self.elseExpression = elseExp

    def traverse(self):
        str = 'if (' + self.condition.traverse() + ') then\n'
        str += '  ' + self.thenExpression.traverse() + '\n'
        str += 'else\n'
        str += '  ' + self.elseExpression.traverse() + '\n'
        str += 'endif \n'
        return str


    


