%{
/* Lexer gets input from a string */

#include <Python.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"
#include "y.tab.h"

#undef YY_INPUT
#define YY_INPUT(b, r, ms) (r = my_yyinput(b, ms))

extern int lineno;
extern void
warning(char *s, char *t);

extern char *strdup(const char *s);
extern int my_yyinput(char *buf, int max);
%}

ws  [ \t\r]+
comment "--".*
qstring \"(\\\"|[^\n\"])*[\n\"]
dot "."
ddot ".."
dcolon  "::"
arrow "->"
atpre "@pre"
nl  \n

%%

{ws}  ;
{comment}  { printf("A line of comment was parsed\n"); }

  /* QSTRING
   * The quoted string should be in one line
   * The exact syntax of a String is undefined in UML1.4 and OCL2.0 for internalization
   */
{qstring}  { if(yytext[yyleng-1] != '"')
             {
               warning("Unterminated string", NULL);
               lineno++;
             }
             yylval.sval = (string)strdup(yytext);
             return QSTRING;
           }

  /* reserved keywords */
inv { return INV; }

pre { return PRE; }

post  { return POST; }

def { return DEF; }

Set { return SET; }

Bag { return BAG; }

Sequence  { return SEQUENCE; }

Tuple { return  TUPLE; }

iterate { return ITERATE; }

[Ll]et { return LET; }

in  { return IN; }

if  { return IF; }

then  { return THEN; }

else  { return ELSE; }

endif { return ENDIF; }

":::" { return TCOLON; }

{dcolon}  { return DCOLON; }

{ddot}  { return DDOT; }

{dot} {return DOT; }

{arrow} { return ARROW; }

{atpre} { return ATPRE; }


  /* reserved relation operations */
"+" { return PLUS; }

"-" { return MINUS; }

"*" { return TIMES; }

"/" { return DIVIDES; }

[Aa][Nn][Dd] { return AND; }

[Oo][Rr]  { return OR; }

[Xx][Oo][Rr] { return XOR; }

[Ii][Mm][Pp][Ll][Ii][Ee][Ss] { return IMPLIES; }

"="   { return EQ; }

"<>"  { return NOTEQ; }

"<="  { return LTEQ; }

">="  { return GTEQ; }

"<"   { return LT; }

">"   { return GT; }


  /* reserved collection operations */
select  { return SELECT; }

collect { return COLLECT; }

reject  { return SYM_REJECT; }

forAll  { return FORALL; }

exists  { return EXISTS; }

isUnique  { return ISUNIQUE; }

sortedBy  { return SORTEDBY; }

any     { return ANY; }

one     { return ONE; }

collectNested { return COLLECTNESTED; }

size    { return SIZE; }

includes { return INCLUDES; }

excludes { return EXCLUDES; }

count   { return COUNT; }

includesAll { return INCLUDESALL; }

excludesAll { return EXCLUDESALL; }

isEmpty { return ISEMPTY; }

notEmpty  { return NOTEMPTY; }

sum { return SUM; }

union   { return UNION; }

intersection { return INTERSECTION; }

including { return INCLUDING; }

excluding { return EXCLUDING; }

symmetricDifference  { return SYMMETRICDIFFERENCE; }

flatten { return FLATTEN; }

asSet { return ASSET; }

asBag { return ASBAG; }

asSequence { return ASSEQUENCE; }

append { return APPEND; }

prepend { return PREPEND; }

subSequence { return SUBSEQUENCE; }

at { return AT; }

first { return FIRST; }

last { return LAST; }

  /* reserved negates operation */
not { return NOT; }



  /* numbers */
[0-9]+  { yylval.ival = atoi(yytext); return INTEGER; }

[0-9]+"."[0-9]* |
"."[0-9]+  |
[0-9]+[eE][-+]?[0-9]+ |
[0-9]+"."[0-9]*[eE][-+]?[0-9]+  |
"."[0-9]+[eE][-+]?[0-9]+  { yylval.dval = atof(yytext); return REAL; }


  /* boolean */
[Tt][Rr][Uu][Ee]  { return SYM_TRUE; }

[Ff][Aa][Ll][Ss][Ee]  { return SYM_FALSE; }


  /* name String
   * The exact syntax of a String is undefined in UML1.4 and OCL2.0 for internalization
   */
[A-Za-z][A-Za-z0-9_]* { yylval.sval = strdup(yytext); return NAMESTRING; }





{nl}  { lineno++;}
.     { return yytext[0]; }

%%
extern char *oclSrc;

static unsigned offset = 0;
static int finish = 0;

/* provide a chunk of stuff to flex */
/* Only one string will be parse */
int
my_yyinput(char *buf, int max)
{
  int len, copylen;
  if (finish == 1){
    finish = 0;
    return 0;  
  }
  len = strlen(oclSrc) - offset;

  if (len >= max)
    copylen = max - 1;
  else
    copylen = len;
  if (len > 0)
    memcpy(buf, oclSrc+offset, copylen);

  if (*(oclSrc+offset+copylen) == '\0') {
    finish = 1;
    offset = 0;
    buf[copylen] = ' ';
    copylen++;
  }
  return copylen;
}

void
yyerror(char *msg)
{
  fprintf(stderr, "At line %d: %s at '%s'\n", lineno, msg, yytext);
}









