%{
/***************************************************************************************
 * This is the third version of the ocl parser
 * Some of the production rules defined in the specification are syntactically
 * ambiguouts, so disambiguating rules have also been defined in the spicification.
 * But those rules may be based on the UML model to which the OCL expression
 * is parsed, thus the UML model must be available when an OCL expression is
 * parsed using those production rules.
 * At the time being, I couldn't figure out how to access the UML model info.
 * So I will change the production rules quite a bit, which means at the end
 * of this first parse, an instance of the abstract syntax could not be build.
 * My final option may be the grammar given in OCL1.4 specification. But I won't
 * chose it as long as I have another choice.
 *
 ***************************************************************************************
 *
 * Notes:
 * 1.The concrete syntax in the OCL2.0 specification does not define
 *   IntegerLiteralExp, RealLiteralExp, StringLiteralExp and BooleanLiteralExp
 *   for internationalization, I make my own decisions here, but they may be
 *   modified later.
 * 2.In the concrete syntax, EnumLiteralExp is defined as:
 *      EnumLiteralExp ::= name '::' <String>
 *   with the disambiguating rule: the specified name must indeed reference
 *   an enumeration. Or, it will have a reduce/reduce conflict with the
 *   production rule for name.
 *   But it is not easy to checked. So, I postpone this question by modifying
 *   the rule as below:
 *       EnumLiteralExp ::=  name ':::' <String>
 * 3.CollectionRange ::= OclExpression '..' OclExpression
 *   Since an OclExpression could be a real number, to make "1..3" parsed correctly,
 *   I decide that "1." is an illegal form of a real number.
 * 4.According to the abstract syntax, the iterator for LoopExp is mandatary.
 *   So, instead of define iterator as:
 *       iterator::= VariableDeclaration?
 *   we could have an iterator or a default iterator, but even by doing that,
 *   the '|' is mandatary, or we will end up with a shift/reduce conflict
 *   I won't use a default_iterator here.
 * 5.In a VarDeclaration, if an initExpression is given, the type must be also given.
 * 6.A referredOperation can be a relational operation, a predefined collection
 *   operation or a operation "mdlProperty" defined in the UML model by the user
 *   "Implies" is added to the "RELOP" list; "not" operation is also added
 *   Furthermore, operationCall, attributeCall and navigationCall will share the keyword
 *   mdlProperty to resolve conflicts. An attributeCall and a navigationCall will be the
 *   same at this stage without knowing the UML model.
 * 7.To parse correctly, the source of the OperationCallExp must be specified
 * 9.The unary minus is not handled at the time being
 *10.OclOperationWithStateArgument operation is not handled at the time being
 *11.The syntax of LetExp is not the same as that in the specification
 *   the semi-colon is specified in the concrete syntax but not in the examples
 *   given in the specification, but i decide to keep the semi-colon
 *12.The precedence of '^' is not given in the concrete syntax, I decide to give it
 *   the same precedence as dot and arrow.
 *   The OclMessageExp need to be studied further.
 *13.TRUE and FALSE will be defined as boolean value on the one hand;
 *   on the other hand, they will be also return from lex as tokens SYM_TRUE and SYM_FALSE
 *14.Token REJECT will be noted as SYM_REJECT, since REJECT is a lex keyword
 */

#include <Python.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "util.h"


extern FILE *yyin, *yyout;

int lineno = 1;

string oclSrc;

extern void yyerror(char *);
extern int yyparse();
extern int yylex();

%}



%union {
  int ival;
  double dval;
  string sval;
  bool bval;
  PyObject *node;
}

%token <ival> INTEGER
%token <dval> REAL
%token <sval> QSTRING NAMESTRING

%token SYM_TRUE SYM_FALSE

%token SELECT COLLECT SYM_REJECT FORALL EXISTS ISUNIQUE
%token SORTEDBY ANY ONE COLLECTNESTED
%token SIZE INCLUDES EXCLUDES COUNT INCLUDESALL EXCLUDESALL
%token ISEMPTY NOTEMPTY SUM
%token UNION INTERSECTION INCLUDING EXCLUDING SYMMETRICDIFFERENCE
%token FLATTEN ASSET ASSEQUENCE ASBAG
%token APPEND PREPEND SUBSEQUENCE AT FIRST LAST
%token NOT

%token INV PRE POST DEF
%token SET BAG SEQUENCE TUPLE
%token DCOLON TCOLON DDOT IN ITERATE ATPRE LET
%token IF THEN ELSE ENDIF

%type <node> OclExpression individualItem body initExpression argument
%type <node> LetExpSub scope condition thenExpression elseExpression
%type <node> PropertyCallExp
%type <node> VariableExp
%type <node> LiteralExp
%type <node> LetExp
%type <node> OclMessageExp
%type <node> IfExp
%type <node> name type
%type <node> EnumLiteralExp
%type <node> CollectionLiteralExp
%type <node> TupleLiteralExp
%type <node> NumericLiteralExp
%type <node> StringLiteralExp
%type <node> BooleanLiteralExp
%type <sval> CollectionTypeIdentifier
%type <node> CollectionItem CollectionItems CollectionItemsPart
%type <node> CollectionRange
%type <node> IntegerLiteralExp;
%type <node> RealLiteralExp;
%type <node> StringLiteralExp;
%type <node> BooleanLiteralExp;
%type <node> TupleLiteralExpParts
%type <node> VariableDeclaration TupleLiteralExpPart iterator nt_result
%type <node> variableDeclarationInLetExp variable
%type <node> ModelPropertyCallExp
%type <node> LoopExp
%type <node> IteratorExp
%type <node> IterateExp
%type <sval> oclStdLibColIter oclStdLibColOp
%type <sval> varName
%type <node> OperationCallExp
%type <node> AttrOrNavCallExp
%type <bval> timeExpr
%type <node> arguments argumentsOrEmpty
%type <sval> mdlProperty
%type <node> variables
%type <node> OclMsgArgOrNull OclMessageArgs OclMessageArg


%left IMPLIES
%left AND OR XOR
%left EQ NOTEQ
%left LT GT LTEQ GTEQ
%left PLUS MINUS
%left TIMES DIVIDES
%left NOT
%left DOT ARROW '^'

%%
start: OclConstraints  { fprintf(yyout, "start\n"); /* setup the initial env of an expression */ }
  ;

OclConstraints: /* empty */
  | OclConstraints OclConstraint  { fprintf(yyout, "OclExpressions\nAt line: %d\n\n",lineno); }
  ;

OclConstraint: constraintType ':' OclExpression { pres = $3; /*printf("## finished parsing\n");*/ }
  ;

constraintType: INV { /*printf("## start to parse\n");*/ }
  | PRE
  | POST
  | DEF
  ;

OclExpression:  PropertyCallExp { $$=$1; fprintf(yyout, "PropertyCallExp\n"); }
  | VariableExp { $$=$1; fprintf(yyout, "VariableExp\n"); }
  | LiteralExp  { $$=$1; fprintf(yyout, "LiteralExp\n"); }
  | LetExp  { $$=$1; fprintf(yyout, "LetExp\n"); }
  | OclMessageExp { $$=$1; fprintf(yyout, "OclMessageExp\n"); }
  | IfExp { $$=$1; fprintf(yyout, "IfExp\n"); }
  | '(' OclExpression ')' { $$=$2; fprintf(yyout, "Parenthesized OclExpression\n"); }
  ;

VariableExp: name {
    fprintf(yyout, "name\n");
    pclass = get_class(pmod, "Py_VariableExp");
    pargs = Py_BuildValue("(O)", $1);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);  
    
    /* Disambiguating rule
     * name must be a name of a visible VaribleDeclaration in the current env 
     * but the check won't be done at this stage */
}
  ;

name: NAMESTRING  {
  fprintf(yyout, "namestring\n"); /* for debugging */
  Py_INCREF(Py_None);
  pclass = get_class(pmod, "Py_Name");
  pargs = Py_BuildValue("(sO)", $1, Py_None);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);  
}
  | NAMESTRING DCOLON name  {
    fprintf(yyout, "name::namestring\n");
    pclass = get_class(pmod, "Py_Name");
    pargs = Py_BuildValue("(sO)", $1, $3);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);  
  }
  ;


  /* define LiteralExp */
LiteralExp: EnumLiteralExp   { $$=$1; fprintf(yyout, "EnumLiteralExp\n"); }
  | CollectionLiteralExp  { $$=$1; fprintf(yyout, "CollectionLiteralExp\n"); }
  | TupleLiteralExp  { $$=$1; fprintf(yyout, "TupleLiteralExp\n"); }
  | NumericLiteralExp  { $$=$1; fprintf(yyout, "NumericLiteralExp\n"); }
  | StringLiteralExp   { $$=$1; fprintf(yyout, "StringLiteralExp\n"); }
  | BooleanLiteralExp  { $$=$1; fprintf(yyout, "BooleanLiteralExp\n"); }
  ;


EnumLiteralExp: name TCOLON NAMESTRING {
  fprintf(yyout, "name||namestring\n");
  pclass = get_class(pmod, "Py_EnumLiteralExp");
  pargs = Py_BuildValue("(Os)", $1, $3);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);  
}
  ;


  /* define CollectionLiteralExp */
CollectionLiteralExp: CollectionTypeIdentifier '{' CollectionItemsPart '}' {
  fprintf(yyout, "CollectionTypeIdentifier '{' CollectionItemsPart '}'\n");
  pclass = get_class(pmod, "Py_CollectionLiteralExp");
  pargs = Py_BuildValue("(sO)", $1, $3);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);  
  
}
  ;

CollectionTypeIdentifier: SET  { 
  $$=(string)strdup("Set");
  fprintf(yyout, "CollectionTypeIdentifier:Set\n"); 
}
  | BAG  { 
  $$=(string)strdup("Bag");
  fprintf(yyout, "CollectionTypeIdentifier:Bag\n"); 
}
  | SEQUENCE  { 
    $$=(string)strdup("Sequence");
    fprintf(yyout, "CollectionTypeIdentifier:Sequence\n"); 
}
  ;

CollectionItemsPart:  /* empty */  { 
    Py_INCREF(Py_None);
    $$ = Py_None; 
    fprintf(yyout, "empty collection\n"); 
}
  | CollectionItems   { $$ = $1; }
  ;

CollectionItems:  CollectionItem  { 
    Py_INCREF(Py_None);
    pclass = get_class(pmod, "Py_CollectionItem");
    pargs = Py_BuildValue("(OO)", $1, Py_None);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs); 
}
  | CollectionItem ',' CollectionItems  { 
    fprintf(yyout, "CollectionItems , CollectionItem\n");
    pclass = get_class(pmod, "Py_CollectionItem");
    pargs = Py_BuildValue("(OO)", $1, $3);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs); 
}
  ;

CollectionItem: CollectionRange  { $$=$1; fprintf(yyout, "CollectionRange\n"); }
  | individualItem  { $$=$1; fprintf(yyout, "IndividualItem\n"); }
  ;

CollectionRange:  OclExpression DDOT OclExpression {
  fprintf(yyout, "OclExp..OclExp\n");
  pclass = get_class(pmod, "Py_CollectionRange");
  pargs = Py_BuildValue("(OO)", $1, $3);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);  
}
  ;

individualItem: OclExpression { $$=$1; fprintf(yyout, "OclExp\n"); }
  ;


  /* define NumericLiteralExp */
NumericLiteralExp:  IntegerLiteralExp  { $$=$1;
                                         fprintf(yyout, "IntegerLiteralExp\n"); }
  | RealLiteralExp  { $$=$1; fprintf(yyout, "RealLiteralExp\n"); }
  ;

IntegerLiteralExp: INTEGER  { 
    fprintf(yyout, "Integer\n");
    pclass = get_class(pmod, "Py_IntegerLiteralExp");
    pargs = Py_BuildValue("(i)", $1);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);      
}
  ;

RealLiteralExp: REAL  {
  fprintf(yyout, "RealExp\n");
  pclass = get_class(pmod, "Py_RealLiteralExp");
  pargs = Py_BuildValue("(d)", $1);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs); 
}
  ;

  /* define StringLiteralExp */
StringLiteralExp: QSTRING  {
  fprintf(yyout, "QStringlExp\n");
  pclass = get_class(pmod, "Py_StringLiteralExp");
  pargs = Py_BuildValue("(s)", $1);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs); 
}
  ;

  /* define BooleanLiteralExp */
BooleanLiteralExp: SYM_TRUE  { 
  fprintf(yyout, "true\n");
  pclass = get_class(pmod, "Py_BooleanLiteralExp");
  pargs = Py_BuildValue("(i)", 1);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);; 
}
  | SYM_FALSE  { 
  fprintf(yyout, "false\n");
  pclass = get_class(pmod, "Py_BooleanLiteralExp");
  pargs = Py_BuildValue("(i)", 0);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs); 
}
  ;


  /* define TupleLiteralExp */
TupleLiteralExp: TUPLE '{' TupleLiteralExpParts '}' {
  fprintf(yyout, "TUPLE '{' non-empty TupleLiteralExpParts '}'\n");
  pclass = get_class(pmod, "Py_TupleLiteralExp");
  pargs = Py_BuildValue("(O)", $3);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);  
}
  | TUPLE '{' '}' {
    pclass = get_class(pmod, "Py_TupleLiteralExp");
    Py_INCREF(Py_None);
    pargs = Py_BuildValue("(O)", Py_None);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);   
}
  ;

TupleLiteralExpParts: TupleLiteralExpPart  {
    pclass = get_class(pmod, "Py_TupleLiteralExpParts");
    Py_INCREF(Py_None);
    pargs = Py_BuildValue("(OO)", $1, Py_None);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);     
}
  | TupleLiteralExpPart ',' TupleLiteralExpParts  {
    pclass = get_class(pmod, "Py_TupleLiteralExpParts");
    Py_INCREF(Py_None);
    pargs = Py_BuildValue("(OO)", $1, $3);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);     
}
  ;

TupleLiteralExpPart:  VariableDeclaration  { $$=$1; fprintf(yyout, "TupleLiteralExpPart\n"); }
  ;


  /* define PropertyCallExp */
PropertyCallExp:  ModelPropertyCallExp  { $$=$1;
                                          fprintf(yyout, "ModelPropertyCallExp\n"); }
  | LoopExp  { $$=$1; fprintf(yyout, "LoopExp\n"); }
  ;

  /* define LoopExp */
LoopExp:  IteratorExp { $$=$1; fprintf(yyout, "IteratorExp\n"); }
  | IterateExp  { $$=$1; fprintf(yyout, "IterateExp\n"); }
  ;

  /* define IteratorExp */
IteratorExp:  OclExpression ARROW oclStdLibColIter '(' iterator '|' body ')' {
  pclass = get_class(pmod, "Py_IteratorExp");
  pargs = Py_BuildValue("(OOOs)", $1, $7, $5, $3);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);
}
  ;

  /* define IterateExp */
IterateExp: OclExpression ARROW ITERATE '(' iterator ';' nt_result '|' body ')' {
  pclass = get_class(pmod, "Py_IterateExp");
  pargs = Py_BuildValue("(OOOO)", $1, $9, $5, $7);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);
}
  ;

oclStdLibColIter:  SELECT  { $$=(string)strdup("select"); fprintf(yyout, "select\n"); }
  | COLLECT { $$=(string)strdup("collect"); fprintf(yyout, "collect\n"); }
  | SYM_REJECT  { $$=(string)strdup("reject"); fprintf(yyout, "reject\n"); }
  | FORALL  { $$=(string)strdup("forAll"); fprintf(yyout, "forAll\n"); }
  | EXISTS  { $$=(string)strdup("exists"); fprintf(yyout, "exists\n"); }
  | ISUNIQUE  { $$=(string)strdup("isUnique"); fprintf(yyout, "isUnique\n"); }
  | SORTEDBY { $$=(string)strdup("sortedBy"); }
  | ANY { $$=(string)strdup("any"); }
  | ONE { $$=(string)strdup("one"); }
  | COLLECTNESTED { $$=(string)strdup("collectNested"); }
  ;

iterator: /* empty */  { 
    Py_INCREF(Py_None);
    $$=Py_None; 
    fprintf(yyout, "default iterator\n"); 
}
  | VariableDeclaration  { $$=$1; fprintf(yyout, "iterator is a VariableDeclaration\n"); }
  ;

nt_result: VariableDeclaration  { $$=$1; fprintf(yyout, "nt_result is a VariableDeclaration\n"); }
  ;

  /* define VariableDeclaration */
VariableDeclaration:  varName ':' type EQ initExpression {
  fprintf(yyout, "VariableDeclaration with type and initExpression\n");
  pclass = get_class(pmod, "Py_VariableDeclaration");
  pargs = Py_BuildValue("(sOO)", $1, $3, $5);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs); 
}
  | varName ':' type  {
    fprintf(yyout, "VariableDeclaration with type but not initExpression\n");
    pclass = get_class(pmod, "Py_VariableDeclaration");
    Py_INCREF(Py_None);   
    pargs = Py_BuildValue("(sOO)", $1, $3, Py_None);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);  
}
  | varName  {
    fprintf(yyout, "VariableDeclaration without type nor initExpression\n");
    pclass = get_class(pmod, "Py_VariableDeclaration");
    Py_INCREF(Py_None);
    Py_INCREF(Py_None);    
    pargs = Py_BuildValue("(sOO)", $1, Py_None, Py_None);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);  
}
  ;

varName:  NAMESTRING  { $$=$1; fprintf(yyout, "VarName is %s\n", $1); }
  ;

type: name { $$=$1; fprintf(yyout, "type is \n"); }
  ;

initExpression: OclExpression  { $$=$1; fprintf(yyout, "initExpression\n"); }
  ;


  /* define ModelPropertyCallExp */
ModelPropertyCallExp: OperationCallExp  { $$=$1;
                                          fprintf(yyout, "OperationCallExp\n"); }
  | AttrOrNavCallExp  { $$=$1;
                        fprintf(yyout, "AttributeCallExp or NavigationCallExp\n"); }
  ;

  /* define OperationCallExp
   * source and arguments are replaced by OclExpression, then conflicts can be solved by
   * precedence
   */
OperationCallExp: OclExpression PLUS OclExpression  {
  fprintf(yyout, "PLUS operation\n");
  pyArgs = make_PyArg($3);
  if (pyArgs == NULL)
    return 1;
  pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "+", 0);
  Py_DECREF(pyArgs);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
  if (pclass == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't get module.class");
    return 1;
  }
  $$ = PyEval_CallObject(pclass, pargs);
  Py_DECREF(pclass);
  Py_DECREF(pargs);
  if ($$ == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
    return 1;
  }  
}
  | OclExpression MINUS OclExpression  {
    fprintf(yyout, "MINUS operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "-", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression TIMES OclExpression  { 
    fprintf(yyout, "TIMES operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "*", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression DIVIDES OclExpression  { 
    fprintf(yyout, "DIVIDES operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "/", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression AND OclExpression  { 
    fprintf(yyout, "AND operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "and", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression OR OclExpression  { 
    fprintf(yyout, "OR operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "or", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression XOR OclExpression  { 
    fprintf(yyout, "XOR operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "xor", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression EQ OclExpression  { 
    fprintf(yyout, "EQ operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "=", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression NOTEQ OclExpression { 
    fprintf(yyout, "NOTEQ operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "<>", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression LTEQ OclExpression  { 
    fprintf(yyout, "MLTEQ operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "<=", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression GTEQ OclExpression  { 
    fprintf(yyout, "GTEQ operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, ">=", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression LT OclExpression  { 
    fprintf(yyout, "LT operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "<", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression GT OclExpression  {
    fprintf(yyout, "GT operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, ">", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression IMPLIES OclExpression  {
    fprintf(yyout, "IMPLIES operation\n");
    pyArgs = make_PyArg($3);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, "implies", 0);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }   
}
  | OclExpression DOT mdlProperty timeExpr '(' argumentsOrEmpty ')' { 
    fprintf(yyout, "model operation\n");
    pyArgs = make_PyArg($6);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, $3, $4);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }    
}
  | OclExpression ARROW oclStdLibColOp timeExpr '(' argumentsOrEmpty ')' {
    fprintf(yyout, "Ocl standard library collection operation\n");
    pyArgs = make_PyArg($6);
    if (pyArgs == NULL)
      return 1;
    pargs = Py_BuildValue("(OOsi)", $1, pyArgs, $3, $4);
    Py_DECREF(pyArgs);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }
}
  | mdlProperty timeExpr '(' argumentsOrEmpty ')'  {
    fprintf(yyout, "model operation with default source\n");
    PyErr_SetString(PyExc_TypeError, "The source of the mdlPropertyCall must be specified");
    return 1;
}
  | NOT OclExpression  { 
    fprintf(yyout, "negates operation\n");
    Py_INCREF(Py_None);
    pargs = Py_BuildValue("(OOsi)", $2, Py_None, "NOT", 0);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    pclass = PyObject_GetAttrString(pmod, "Py_OperationCallExp");
    if (pclass == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't get module.class");
      return 1;
    }
    $$ = PyEval_CallObject(pclass, pargs);
    Py_DECREF(pclass);
    Py_DECREF(pargs);
    if ($$ == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
      return 1;
    }     
}
  ;

oclStdLibColOp: SIZE { $$ = (string)strdup("size"); }
  | INCLUDES { $$ = (string)strdup("includes"); }
  | EXCLUDES { $$ = (string)strdup("excludes"); }
  | COUNT { $$ = (string)strdup("count"); }
  | INCLUDESALL { $$ = (string)strdup("includesAll"); }
  | EXCLUDESALL { $$ = (string)strdup("excludesAll"); }
  | ISEMPTY { $$ = (string)strdup("isEmpty"); }
  | NOTEMPTY { $$ = (string)strdup("notEmpty"); }
  | SUM { $$ = (string)strdup("sum"); }
  | UNION { $$ = (string)strdup("union"); }
  | INTERSECTION { $$ = (string)strdup("intersection"); }
  | INCLUDING { $$ = (string)strdup("including"); }
  | EXCLUDING { $$ = (string)strdup("excluding"); }
  | SYMMETRICDIFFERENCE { $$ = (string)strdup("symmetricDifference"); }
  | FLATTEN { $$ = (string)strdup("flatten"); }
  | ASSET { $$ = (string)strdup("asSet"); }
  | ASBAG { $$ = (string)strdup("asBag"); }
  | ASSEQUENCE { $$ = (string)strdup("asSequence"); }
  | APPEND { $$ = (string)strdup("append"); }
  | PREPEND { $$ = (string)strdup("prepend"); }
  | SUBSEQUENCE { $$ = (string)strdup("subSequence"); }
  | AT { $$ = (string)strdup("at"); }
  | FIRST { $$ = (string)strdup("first"); }
  | LAST { $$ = (string)strdup("last"); }
  ;

timeExpr:  /* empty */  { $$=0; }
  | ATPRE  { $$=1; fprintf(yyout, "at Pre\n"); }
  ;

argumentsOrEmpty:  /* empty */  {
    Py_INCREF(Py_None);
    $$=Py_None;
    fprintf(yyout, "empty argument\n"); 
}
  | arguments { $$=$1; fprintf(yyout, "non empty argumentsn\n"); }
  ;


argument: OclExpression  { $$=$1; fprintf(yyout, "argument\n"); }
  ;

arguments: argument {
  pclass = get_class(pmod, "Py_Arguments");
  Py_INCREF(Py_None);
  pargs = Py_BuildValue("(OO)", $1, Py_None);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);  
}
  | argument ',' arguments  {
    pclass = get_class(pmod, "Py_Arguments");
      pargs = Py_BuildValue("(OO)", $1, $3);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);  
}
  ;


body: OclExpression { $$ = $1; fprintf(yyout, "body\n"); }
  ;

mdlProperty:  NAMESTRING  { $$=$1; fprintf(yyout, "mdlProperty: %s\n", $1); }
  ;

  /* define AttrOrNavCallExp
   * we can only know if it is an AttributeCallExp or a NavigationCallExp
   * when we know the UML model, so this will be done in the next pass
   * source must be specified to resolve the reduce/reduce conflicts
   * thus, "AttrOrNavCallExp::=mdlProperty timeExpr" is not valid
   */
AttrOrNavCallExp: OclExpression DOT mdlProperty timeExpr {
  fprintf(yyout, "AttrOrNavCallExp with source\n");
  pargs = Py_BuildValue("(Osi)", $1, $3, $4);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  pclass = PyObject_GetAttrString(pmod, "Py_AttrOrNavCallExp");
  if (pclass == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't get module.class");
    return 1;
  }
  $$ = PyEval_CallObject(pclass, pargs);
  Py_DECREF(pclass);
  Py_DECREF(pargs);
  if ($$ == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
    return 1;
  }      
}
  ;


  /* define LetExp
   * LetExp will not produce an instance of abstract syntax
   * but it will modify the env
   */
LetExp: LET variables LetExpSub {
  pclass = get_class(pmod, "Py_LetExp");
  pargs = Py_BuildValue("(OO)", $2, $3);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs); 
}
  ;

variables:  variable { 
  pclass = get_class(pmod, "Py_Variables");
  Py_INCREF(Py_None);
  pargs = Py_BuildValue("(OO)", $1, Py_None);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);
}
  | variable ';' variables  {
    fprintf(yyout, "multi-var def\n"); 
    pclass = get_class(pmod, "Py_Variables");
    pargs = Py_BuildValue("(OO)", $1, $3);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs); 
}
  ;

variable: variableDeclarationInLetExp { $$=$1; }
  ;

variableDeclarationInLetExp: varName ':' type EQ initExpression {
  fprintf(yyout, "variableDeclarationInLetExp with type and initExpression\n");
  pclass = get_class(pmod, "Py_VariableDeclaration");
  pargs = Py_BuildValue("(sOO)", $1, $3, $5);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);  
}
  ;

LetExpSub:  IN scope  { $$=$2; fprintf(yyout, "LetExpSub\n"); }
  ;

  /* parenthesis is added to resolve conflicts
   * but it's not the same as the grammar given in the specification
   */
scope:  '(' OclExpression ')'  { $$=$2; fprintf(yyout, "scope of the let expr\n"); }
  ;


  /* define OclMessageExp
   * It need to be studied further
   */
OclMessageExp:  OclExpression '^' NAMESTRING '(' OclMsgArgOrNull ')' {
  pclass = get_class(pmod, "Py_OclMessageExp");
  pargs = Py_BuildValue("(OsO)", $1, $3, $5);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);      
}
  ;

OclMsgArgOrNull:  /* empty */ { 
  fprintf(yyout, "no args\n");
  Py_INCREF(Py_None);
  $$ = Py_None;
}
  | OclMessageArgs { $$=$1; fprintf(yyout, "OclMsgArg\n"); }
  ;

OclMessageArgs:  OclMessageArg {
    pclass = get_class(pmod, "Py_OclMessageArg");
    Py_INCREF(Py_None);
    pargs = Py_BuildValue("(OO)", $1, Py_None);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);    
}
  | OclMessageArg ',' OclMessageArgs {
    pclass = get_class(pmod, "Py_OclMessageArg");
    pargs = Py_BuildValue("(OO)", $1, $3);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);     
}
  ;

OclMessageArg:  '?' {
    pclass = get_class(pmod, "Py_UnspecifiedValueExp");
    Py_INCREF(Py_None);
    pargs = Py_BuildValue("(O)", Py_None);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);    
}
  | '?' ':' type  {
    pclass = get_class(pmod, "Py_UnspecifiedValueExp");
    pargs = Py_BuildValue("(O)", $3);
    if (pargs == NULL){
      PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
      return 1;
    }
    $$ = get_inst(pclass, pargs);     
}
  | OclExpression  { $$=$1; }
  ;


  /* define IfExp
   */
IfExp:  IF condition THEN thenExpression ELSE elseExpression ENDIF {
  pclass = get_class(pmod, "Py_IfExp");
  pargs = Py_BuildValue("(OOO)", $2, $4, $6);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return 1;
  }
  $$ = get_inst(pclass, pargs);   
}
  ;

condition: OclExpression { $$ = $1; }
  ;

thenExpression: OclExpression { $$ = $1; }
  ;

elseExpression: OclExpression { $$ = $1; }
  ;



%%
string progname = "ocl_Parse";

#define DEFAULT_TMPCHECK "screen.out"

string usage = "usage: %s [oclSource] [tmpCheck]\n";

PyObject *pmod, *pclass, *pemptyArg, *pargs, *pyArgs, *pres;
/*
PyObject *py_variableExp, *py_name, *py_emumLiteralExp;
PyObject *py_collectionLiteralExp, *py_collectionItem, *py_collctionRange;
PyObject *py_integerLiteralExp, *py_realLiteralExp, *py_stringLiteralExp;
PyObject *py_booleanLiteralExp, *py_tupleLiteralExp;
PyObject *py_tupleLiteralExpParts, *py_iteratorExp, *py_iterateExp;
PyObject *py_variableDeclaration, *py_operationCallExp;
PyObject *py_attrOrNavCallExp, *py_arguments, *py_letExp;
PyObject *py_variables, *py_oclMessageExp, *py_oclMessageArg;
PyObject *py_unspecifiedValueExp, *py_ifExp;
*/

/* modified */
static PyObject *
ocl_Parse(PyObject *self, PyObject *args)
{
  string ocl_src = NULL;
  string tmpCheck = NULL;

  
  /* Initialize the Python constructor objects */
  pmod = PyImport_ImportModule("absyn");  /* fetch "absyn.py" module */
  if (pmod == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't fetch absyn.py");
    return NULL;
  }

  pemptyArg = Py_BuildValue("()");
  if (pemptyArg == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return NULL;
  }

  /* parse the arguments recieved from python caller */
  if (! PyArg_Parse(args, "(s)", &ocl_src))
    return NULL;

  /* input from a string */
  oclSrc = ocl_src;

  if (tmpCheck == NULL)
    {
      tmpCheck = strdup(DEFAULT_TMPCHECK);
      yyout = fopen (tmpCheck, "w");
      assert(yyout);
    }

  /* start to creat python file for the AST in ATOM3 */

  /* start to parse the Ocl sourse file */
  if (yyparse() == 1){
    PyErr_SetString(PyExc_TypeError, "Error in Lexing");
    return NULL;
  }

  /* finished parsing */
  return pres;
}

static struct PyMethodDef ocl_methods[] = {
  {"ocl_Parse", ocl_Parse, 1},
  {NULL, NULL}
};

void initocl()
{
  (void) Py_InitModule("ocl", ocl_methods);
}

void
warning(string s, string t) /* print warning message */
{
  fprintf(stderr, "%s: %s", progname, s);
  if (t)
    fprintf(stderr, " %s", t);
  fprintf(stderr, " line %d\n", lineno);
}

/* helper function to fetch a python class */
static PyObject *
get_class(PyObject *pmod, string class_name)
{
  PyObject *pclass;
  
  /* fetch module.class */
  pclass = PyObject_GetAttrString(pmod, class_name);
  if (pclass == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't get module.class");
    exit(1);
    /*return NULL;*/
  }
  return pclass;
}

/* helper function to instantiate an instance */
static PyObject *
get_inst(PyObject *pclass, PyObject *pargs)
{
  PyObject *pinst;
  pinst = PyEval_CallObject(pclass, pargs);
  Py_DECREF(pclass);
  Py_DECREF(pargs);
  if (pinst == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't call module.class()");
    exit(1);
    /*return NULL;*/
  }
  return pinst;
}

/* helper function to creat an Py_Arguments */
static PyObject *
make_PyArg(PyObject *oclExp)
{
  PyObject *py_args;
  pclass = PyObject_GetAttrString(pmod, "Py_Arguments");
  if (pclass == NULL){
    PyErr_SetString(PyExc_TypeError, "Cannot get module.class");
    return NULL;
  }
  Py_INCREF(Py_None);
  pargs = Py_BuildValue("(OO)", oclExp, Py_None);
  if (pargs == NULL){
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
    return NULL;
  }
  py_args = PyEval_CallObject(pclass, pargs);
  Py_DECREF(pclass);
  Py_DECREF(pargs);
  if (py_args == NULL)
    PyErr_SetString(PyExc_TypeError, "Can't build arguments list");
  return py_args;
}




































